/*
    $Id: bsf.c,v 1.3 1999/07/11 01:51:06 james Exp $

    bsf, Videopad Object Filing System
    Copyright (C) 1999  James Cameron  (quozl@us.netrek.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef lint
static char vcid[] = "$Id: bsf.c,v 1.3 1999/07/11 01:51:06 james Exp $";
#endif /* lint */

/*
    Provides back end database support for the filing system application
    on the Videopad.  Maintains an in-memory database of the object data,
    with a transaction journal on disk.  The in-memory database is loaded
    from the journal on initialisation.
*/

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "bsf.h"

#define MAXNAME 80
#define MAXREPLY 60
#define MAXLINE 80
#define MAXMSG 2048
#define MAXITEMS 50

struct item
{
    struct item *next;
    int no;
    char name[MAXNAME];
    struct item *parent;	/* can be NULL */
};

struct item *items;

FILE *transactions;

int last;

int updated;
int touched;

char response[MAXMSG];

struct item *bsf_locate ( int no )
{
    struct item *item = items;
    
    while ( item != NULL )
    {
        if ( item->no == no ) return item;
        item = item->next;
    }
    return NULL;
}

char *bsf_store ( ) /* "||store|%d|%s", parent-number, item-name */
{
    int no;
    char *name;
    struct item *item, *parent;
    
    no = atoi ( strtok ( NULL, "|" ) );
    
    if ( no == 0 )
    {
        parent = NULL;
    }
    else
    {
        parent = bsf_locate ( no );
        if ( parent == NULL )
        {
            static char message[MAXREPLY];
            
            sprintf ( message, "Can't store item, parent|entered (%d) not known", no );
            return message;
        }
    }
    
    name = strtok ( NULL, "|" );
    
    item = malloc ( sizeof ( struct item ) );
    item->next = items;
    item->no = ++last;
    strcpy ( item->name, name );
    item->parent = parent;
    items = item;
    updated++;
    
    sprintf ( response, "||stored|%d", item->no );
    return response;
}

char *bsf_rename ( ) /* "||rename|%d|%s", number, string */
{
    int no;
    char *name;
    struct item *item;
    
    no = atoi ( strtok ( NULL, "|" ) );
    item = bsf_locate ( no );
    if ( item == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't rename item,\nno such item %d", no );
        return message;
    }
    
    strcpy ( item->name, strtok ( NULL, "|" ) );
    updated++;
    return "||renamed";
}

char *bsf_enclose ( ) /* "||enclose|%d|%d", parent, item */
{
    int no;
    struct item *item, *parent;
    
    no = atoi ( strtok ( NULL, "|" ) );
    parent = bsf_locate ( no );
    if ( parent == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't move item,\nno such parent %d", no );
        return message;
    }
    
    no = atoi ( strtok ( NULL, "|" ) );
    item = bsf_locate ( no );
    if ( item == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't move item,\nno such item %d", no );
        return message;
    }
    
    item->parent = parent;
    updated++;
    return "||enclosed";
}

char *bsf_move ( ) /* "||move|%d|%d", number, location */
{
    int no;
    struct item *item, *parent;
    
    no = atoi ( strtok ( NULL, "|" ) );
    item = bsf_locate ( no );
    if ( item == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't move item,\nno such item %d", no );
        return message;
    }
    
    no = atoi ( strtok ( NULL, "|" ) );
    parent = bsf_locate ( no );
    if ( parent == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't move item,\nno such parent %d", no );
        return message;
    }
    
    item->parent = parent;
    updated++;
    return "||moved";
}

char *bsf_contain ( ) /* "||contain|%d", number */
{
    int no;
    int i;
    struct item *item, *parent;
    
    no = atoi ( strtok ( NULL, "|" ) );
    parent = bsf_locate ( no );
    if ( parent == NULL )
    {
        static char message[MAXREPLY];
        
        sprintf ( message, "Can't contain,\nno parent %d", no );
        return message;
    }
    
    sprintf ( response, "||contained|1" );
    
    item = items;
    i = 0;
    while ( item != NULL )
    {
        if ( item->parent == parent )
        {
            char buffer[MAXLINE];
            
            sprintf ( buffer, "|%d|%.*s", item->no, MAXNAME, item->name );
            strcat ( response, buffer );
            i++;
            if ( i == MAXITEMS ) break;
        }
        item = item->next;
    }
    
    if ( i == 0 )
    {
        sprintf ( response, "Item %d is childless", no );
        return response;
    }
    
    return response;
}

char *bsf_find ( )
{
    char *find;
    int no;
    struct item *item;
    int i;
    
    find = strtok ( NULL, "|" );
    if ( find == NULL ) return "malformed request";
    
    sprintf ( response, "||found|1" );
    
    item = items;
    i = 0;
    while ( item != NULL )
    {
        if ( strstr ( item->name, find ) != NULL )
        {
            char buffer[MAXLINE];
            
            sprintf ( buffer, "|%d|%.*s", item->no, MAXNAME, item->name );
            strcat ( response, buffer );
            i++;
            if ( i == MAXITEMS ) break;
        }
        item = item->next;
    }
    
    if ( i == 0 )
    {
        sprintf ( response, "Found nothing for %s", find );
        return response;
    }
    
    return response;
}

char *bsf_show ( )
{
    int no, i;
    struct item *item, *parent, *child;
    
    no = atoi ( strtok ( NULL, "|" ) );
    item = bsf_locate ( no );
    if ( item == NULL ) return "No such item";
    
    child = items;
    while ( child != NULL )
    {
        if ( child->parent == item ) break;
        child = child->next;
    }
    
    parent = item->parent;
    if ( parent == NULL )
    {
        sprintf ( response, "||showed|1|%.*s|%d|%.*s|%d",
            MAXNAME, item->name, 0, MAXNAME, "(root)", child == 0 ? 0 : 1 );
    }
    else
    {
        sprintf ( response, "||showed|1|%.*s|%d|%.*s|%d",
            MAXNAME, item->name, parent->no,
            MAXNAME, parent->name, child == 0 ? 0 : 1 );
    }
        
    return response;
}

char *bsf_action ( char *pdu )
{
    char *token;
    char message[MAXMSG];
    
    strcpy ( message, pdu );
    token = strtok ( message, "|" );
    
    if ( ! strcmp ( token, "find"    ) ) return bsf_find();
    if ( ! strcmp ( token, "show"    ) ) return bsf_show();
    if ( ! strcmp ( token, "contain" ) ) return bsf_contain();
    
    if ( ! strcmp ( token, "move"    ) ) return bsf_move();
    if ( ! strcmp ( token, "enclose" ) ) return bsf_enclose();
    if ( ! strcmp ( token, "rename"  ) ) return bsf_rename();
    if ( ! strcmp ( token, "store"   ) ) return bsf_store();
    
    return NULL;
}

char *bsf ( int rlun, const char *buffer, int length )
{
    char *token;
    char message[2048], *pdu;
    
    memcpy ( message, buffer, length );
    message[length] = '\0';
    pdu = message;
    
    printf ( "%s\n", pdu );
    
    if ( *pdu++ != '|' ) return NULL;
    if ( *pdu++ != '|' ) return NULL;
    if ( strncmp ( pdu, "filing", 6 ) ) return NULL;
    
    pdu += 7;
    
    updated = 0;
    token = bsf_action ( pdu );
    if ( updated )
    {
        touched++;
        fprintf ( transactions, "%s\n", pdu );
        printf ( "Logged: %s\n", pdu );
    }
    
    if ( token != NULL )
    {
        printf ( "%s\n", token );
        return token;
    }
    return NULL;
}

void bsf_load ( )
{
    FILE *file;
    char buffer[MAXLINE];
    
    touched = 0;
    items = NULL;
    last = 0;
    
    file = fopen ( "bsf.dat", "r" );
    if ( file == NULL ) {
	if (errno != ENOENT) {
            perror ( "bsf_load: fopen" );
	    exit(0);
	}
    } else {
        while(1) {
	    if ( feof ( file ) ) break;
	    fgets ( buffer, MAXLINE, file );
	    if ( feof ( file ) ) break;
	    buffer[strlen(buffer)-1] = '\0';
	    bsf_action ( buffer );
	}
	fclose ( file );
    }

    transactions = fopen ( "bsf.dat", "a" );
    if ( transactions == NULL )
    {
        perror ( "bsf_load: fopen" );
        exit(0);
    }
}

bsf_logoff ( unsigned char lun )
{
    int stat;
    
    if ( touched == 0 ) return;
    touched = 0;
    
    stat = fclose ( transactions );
    if ( stat < 0 )
    {
        perror ( "bsf_logoff: fclose" );
        return;
    }
    transactions = fopen ( "bsf.dat", "a" );
    if ( transactions == NULL )
    {
        perror ( "bsf_logoff: fopen" );
        exit(0);
    }
    printf ( "bsf_logoff: updates flushed to disk\n" );
}

bsf_logon ( char *id, unsigned char lun )
{
}
